import logging

import pandas as pd

from lactate_thresholds import determine
from lactate_thresholds.zones import (
    friel_7_zones_running,
    seiler_3_zones,
    seiler_5_zones,
)

col_set = ["zone", "intensity", "heart_rate", "focus"]


def test_seiler_zones(test_instances):
    df = pd.DataFrame.from_dict(test_instances["cycling2"])
    r = determine(df, lactate_col="lactate_8")

    zones = seiler_3_zones(r)
    assert list(zones.columns) == col_set

    logging.info(zones)
    zones = seiler_5_zones(r)
    assert list(zones.columns) == col_set
    logging.info(zones)


def test_friel_zones(test_instances):
    df = pd.DataFrame.from_dict(test_instances["cycling2"])
    r = determine(df, lactate_col="lactate_8")

    zones = friel_7_zones_running(r)
    assert list(zones.columns) == col_set
    logging.info(zones)



import base64
import json
import os
from io import StringIO

import pandas as pd
import streamlit as st

import lactate_thresholds as lt
import lactate_thresholds.zones as zones


def get_base_url():
    import urllib.parse

    session = st.runtime.get_instance()._session_mgr.list_active_sessions()[0]
    st_base_url = urllib.parse.urlunparse(
        [session.client.request.protocol, session.client.request.host, "", "", "", ""]
    )
    return st_base_url


def data_placeholder() -> pd.DataFrame:
    df = pd.DataFrame.from_dict(
        [
            {
                "step": 1,
                "length": 8,
                "intensity": 100,
                "rel_power": 1.3,
                "heart_rate": 113,
                "lactate_4": 1.0,
                "lactate_8": 1.0,
                "cadence": 102,
                "rpe": 6,
            },
            {
                "step": 2,
                "length": 8,
                "intensity": 140,
                "rel_power": 1.8,
                "heart_rate": 126,
                "lactate_4": 1.0,
                "lactate_8": 1.0,
                "cadence": 100,
                "rpe": 7,
            },
            {
                "step": 3,
                "length": 8,
                "intensity": 180,
                "rel_power": 2.3,
                "heart_rate": 137,
                "lactate_4": 0.9,
                "lactate_8": 0.9,
                "cadence": 100,
                "rpe": 10,
            },
            {
                "step": 4,
                "length": 8,
                "intensity": 220,
                "rel_power": 2.8,
                "heart_rate": 151,
                "lactate_4": 1.0,
                "lactate_8": 1.0,
                "cadence": 98,
                "rpe": 12,
            },
            {
                "step": 5,
                "length": 8,
                "intensity": 260,
                "rel_power": 3.3,
                "heart_rate": 168,
                "lactate_4": 1.9,
                "lactate_8": 1.9,
                "cadence": 98,
                "rpe": 16,
            },
            {
                "step": 6,
                "length": 8,
                "intensity": 300,
                "rel_power": 3.8,
                "heart_rate": 181,
                "lactate_4": 3.3,
                "lactate_8": 3.8,
                "cadence": 94,
                "rpe": 18,
            },
            {
                "step": 7,
                "length": 8,
                "intensity": 340,
                "rel_power": 4.3,
                "heart_rate": 190,
                "lactate_4": 6.4,
                "lactate_8": 7.5,
                "cadence": 92,
                "rpe": 19,
            },
        ]
    )
    return lt.clean_data(df, lactate_col="lactate_8")


def main():
    st.set_page_config(
        page_title="Lactate Thresholds",
        page_icon="🌟",
        layout="wide",
        initial_sidebar_state="auto",
    )

    def snapshot_url():
        snapshot = {
            "measurements": df_editor.to_json(index=False, orient="records"),
            "lt1": st.session_state.lt_df.loc[0, "intensity"],
            "lt2": st.session_state.lt_df.loc[1, "intensity"],
            "zone_type": st.session_state.zone_type,
        }

        json_snapshot = json.dumps(snapshot)
        base64_str = base64.b64encode(json_snapshot.encode("utf-8")).decode("utf-8")

        st.session_state.snapshot_url = f"{get_base_url()}/?snapshot={base64_str}"

    @st.cache_data
    def init_measurements_df(snapshop_b64: str = None) -> pd.DataFrame:
        if snapshop_b64:
            # try to load snapshot
            try:
                snapshot = json.loads(base64.b64decode(st.query_params["snapshot"]).decode("utf-8"))
                df = pd.read_json(StringIO(snapshot["measurements"]))
                st.session_state.lt1_setting = snapshot["lt1"]
                st.session_state.lt2_setting = snapshot["lt2"]
                st.session_state.zone_type = snapshot["zone_type"]
                return df
            except Exception as e:
                st.warning(f"Error loading snapshot: {e}")

        return data_placeholder()

    df = init_measurements_df(st.query_params.get("snapshot"))
    st.title("Lactate Thresholds")

    df_editor = st.data_editor(
        df,
        num_rows="dynamic",
        use_container_width=True,
        column_config={
            "intensity": st.column_config.NumberColumn("intensity", step=0.1),
        },
    )
    results = lt.determine(df_editor)

    def construct_lt_df():
        lt_df = pd.DataFrame([results.lt1_estimate.model_dump(), results.lt2_estimate.model_dump()])
        lt_df.insert(0, "Threshold", ["LT1", "LT2"])
        st.session_state.lt_df = lt_df

    def construct_zones_df():
        if st.session_state.zone_type == "Seiler 3-zone":
            zones_df = zones.seiler_3_zones(results)

        elif st.session_state.zone_type == "Seiler 5-zone":
            zones_df = zones.seiler_5_zones(results)

        elif st.session_state.zone_type == "Friel 7-zone":
            zones_df = zones.friel_7_zones_running(results)

        else:
            zones_df = pd.DataFrame()

        st.session_state.zones_df = zones_df

    def update_lt():
        results.calc_lt1_lt2_estimates(lt1=st.session_state.lt1_setting, lt2=st.session_state.lt2_setting)
        construct_lt_df()
        construct_zones_df()

    hcol1, hcol2 = st.columns([0.7, 0.3])
    with hcol1:
        st.checkbox("Show fit line", key="fit_line", value=True)
        st.altair_chart(
            lt.plot.lactate_intensity_plot(results, show_fit_line=st.session_state.fit_line),
            use_container_width=True,
        )

        st.altair_chart(
            lt.plot.heart_rate_intensity_plot(results, show_fit_line=st.session_state.fit_line),
            use_container_width=True,
        )

    if "lt_df" not in st.session_state:
        construct_lt_df()

    with hcol2:
        st.markdown("**Set LT1 and LT2 intensity values**")
        st.markdown(":gray[Defaults to estimated values]")

        if "lt1_setting" not in st.session_state:
            st.session_state.lt1_setting = st.session_state.lt_df.loc[0, "intensity"]

        if "lt2_setting" not in st.session_state:
            st.session_state.lt2_setting = st.session_state.lt_df.loc[1, "intensity"]

        col1, col2 = st.columns(2)
        with col1:
            st.number_input(
                "LT1",
                on_change=update_lt,
                key="lt1_setting",
            )
        with col2:
            st.number_input(
                "LT2",
                on_change=update_lt,
                key="lt2_setting",
            )

        st.dataframe(st.session_state.lt_df, hide_index=True, use_container_width=True)

        st.selectbox(
            "Select zones type",
            ["Seiler 3-zone", "Seiler 5-zone", "Friel 7-zone"],
            key="zone_type",
            on_change=construct_zones_df,
        )

        if "zones_df" not in st.session_state:
            construct_zones_df()

        st.dataframe(st.session_state.zones_df, hide_index=True, use_container_width=True)

        with st.popover("Link to snapshot"):
            snapshot_url()
            st.code(st.session_state.snapshot_url)


def start():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    import subprocess
    import sys

    subprocess.run(["streamlit", "run", f"{current_dir}/app.py"] + sys.argv[1:])


if __name__ == "__main__":
    main()



import os

import pytest


@pytest.fixture
def test_output_dir():
    """Ensure the test_output directory exists, or create it."""
    dir_path = "test_output"

    if not os.path.exists(dir_path):
        os.makedirs(dir_path)

    assert os.path.isdir(dir_path), f"{dir_path} is not a directory"

    return dir_path


@pytest.fixture
def test_instances():
    return {
        "simple": [
            {"step": 0, "length": 0, "intensity": 0, "lactate": 1.0, "heart_rate": 80},
            {
                "step": 1,
                "length": 4,
                "intensity": 8.5,
                "lactate": 1.2,
                "heart_rate": 122,
            },
            {
                "step": 2,
                "length": 4,
                "intensity": 10.0,
                "lactate": 1.3,
                "heart_rate": 131,
            },
            {
                "step": 3,
                "length": 4,
                "intensity": 11.5,
                "lactate": 1.0,
                "heart_rate": 145,
            },
            {
                "step": 4,
                "length": 4,
                "intensity": 13.0,
                "lactate": 0.8,
                "heart_rate": 155,
            },
            {
                "step": 5,
                "length": 4,
                "intensity": 14.5,
                "lactate": 1.1,
                "heart_rate": 162,
            },
            {
                "step": 6,
                "length": 4,
                "intensity": 16.0,
                "lactate": 1.8,
                "heart_rate": 171,
            },
            {
                "step": 7,
                "length": 4,
                "intensity": 17.5,
                "lactate": 2.9,
                "heart_rate": 180,
            },
            {
                "step": 8,
                "length": 4,
                "intensity": 19.0,
                "lactate": 5.3,
                "heart_rate": 186,
            },
        ],
        "differently_named_cols": [
            {
                "step": 0,
                "step_length": 0,
                "intensity": 0,
                "lactate": 1.0,
                "heartrate": 80,
            },
            {
                "step": 1,
                "step_length": 4,
                "intensity": 8.5,
                "lactate": 2.2,
                "heartrate": 122,
            },
            {
                "step": 2,
                "step_length": 4,
                "intensity": 10.0,
                "lactate": 1.3,
                "heartrate": 131,
            },
            {
                "step": 3,
                "step_length": 4,
                "intensity": 11.5,
                "lactate": 1.0,
                "heartrate": 145,
            },
            {
                "step": 4,
                "step_length": 4,
                "intensity": 13.0,
                "lactate": 0.8,
                "heartrate": 155,
            },
            {
                "step": 5,
                "step_length": 4,
                "intensity": 14.5,
                "lactate": 1.1,
                "heartrate": 162,
            },
            {
                "step": 6,
                "step_length": 4,
                "intensity": 16.0,
                "lactate": 1.8,
                "heartrate": 171,
            },
            {
                "step": 7,
                "step_length": 4,
                "intensity": 17.5,
                "lactate": 2.9,
                "heartrate": 180,
            },
            {
                "step": 8,
                "step_length": 4,
                "intensity": 19.0,
                "lactate": 5.3,
                "heartrate": 186,
            },
        ],
        "non_numeric_vals": [
            {"step": 0, "length": 0, "intensity": 0, "lactate": 1.0, "heartrate": 80},
            {
                "step": 1,
                "length": 4,
                "intensity": 8.5,
                "lactate": 2.2,
                "heartrate": 122,
            },
            {
                "step": 2,
                "length": 4,
                "intensity": "a",
                "lactate": 1.3,
                "heartrate": 131,
            },
            {
                "step": 3,
                "length": 4,
                "intensity": 11.5,
                "lactate": 1.0,
                "heartrate": 145,
            },
            {
                "step": 4,
                "length": 4,
                "intensity": 13.0,
                "lactate": 0.8,
                "heartrate": 155,
            },
            {
                "step": 5,
                "length": 4,
                "intensity": 14.5,
                "lactate": 1.1,
                "heartrate": 162,
            },
            {
                "step": 6,
                "length": 4,
                "intensity": 16.0,
                "lactate": 1.8,
                "heartrate": 171,
            },
            {
                "step": 7,
                "length": 4,
                "intensity": 17.5,
                "lactate": 2.9,
                "heartrate": 180,
            },
            {
                "step": 8,
                "length": 4,
                "intensity": 19.0,
                "lactate": 5.3,
                "heartrate": 186,
            },
        ],
        "cycling1": [
            {
                "step": 1,
                "length": 8,
                "intensity": 100,
                "rel_power": 1.2,
                "heart_rate": 120,
                "lactate_4": 1.6,
                "lactate_8": 1.6,
                "cadence": 98,
                "rpe": 6,
                "feeling": "Zeer zeer licht",
            },
            {
                "step": 2,
                "length": 8,
                "intensity": 140,
                "rel_power": 1.7,
                "heart_rate": 132,
                "lactate_4": 1.3,
                "lactate_8": 1.3,
                "cadence": 97,
                "rpe": 9,
                "feeling": "Zeer licht",
            },
            {
                "step": 3,
                "length": 8,
                "intensity": 180,
                "rel_power": 2.2,
                "heart_rate": 144,
                "lactate_4": 1.4,
                "lactate_8": 1.4,
                "cadence": 96,
                "rpe": 11,
                "feeling": "Tamelijk licht",
            },
            {
                "step": 4,
                "length": 8,
                "intensity": 220,
                "rel_power": 2.7,
                "heart_rate": 160,
                "lactate_4": 1.8,
                "lactate_8": 1.8,
                "cadence": 91,
                "rpe": 14,
                "feeling": "Redelijk zwaar",
            },
            {
                "step": 5,
                "length": 8,
                "intensity": 260,
                "rel_power": 3.2,
                "heart_rate": 176,
                "lactate_4": 2.9,
                "lactate_8": 3.4,
                "cadence": 81,
                "rpe": 16,
                "feeling": "Zwaar",
            },
            {
                "step": 6,
                "length": 8,
                "intensity": 300,
                "rel_power": 3.7,
                "heart_rate": 192,
                "lactate_4": 5.5,
                "lactate_8": 6.2,
                "cadence": 81,
                "rpe": 19,
                "feeling": "Zeer zeer zwaar",
            },
        ],
        "cycling2": [
            {
                "step": 1,
                "length": 8,
                "intensity": 100,
                "rel_power": 1.3,
                "heart_rate": 113,
                "lactate_4": 1.0,
                "lactate_8": 1.0,
                "cadence": 102,
                "rpe": 6,
                "feeling": "Zeer zeer licht",
            },
            {
                "step": 2,
                "length": 8,
                "intensity": 140,
                "rel_power": 1.8,
                "heart_rate": 126,
                "lactate_4": 1.0,
                "lactate_8": 1.0,
                "cadence": 100,
                "rpe": 7,
                "feeling": "Zeer zeer licht",
            },
            {
                "step": 3,
                "length": 8,
                "intensity": 180,
                "rel_power": 2.3,
                "heart_rate": 137,
                "lactate_4": 0.9,
                "lactate_8": 0.9,
                "cadence": 100,
                "rpe": 10,
                "feeling": "Zeer licht",
            },
            {
                "step": 4,
                "length": 8,
                "intensity": 220,
                "rel_power": 2.8,
                "heart_rate": 151,
                "lactate_4": 1.0,
                "lactate_8": 1.0,
                "cadence": 98,
                "rpe": 12,
                "feeling": "Tamelijk licht",
            },
            {
                "step": 5,
                "length": 8,
                "intensity": 260,
                "rel_power": 3.3,
                "heart_rate": 168,
                "lactate_4": 1.9,
                "lactate_8": 1.9,
                "cadence": 98,
                "rpe": 16,
                "feeling": "Zwaar",
            },
            {
                "step": 6,
                "length": 8,
                "intensity": 300,
                "rel_power": 3.8,
                "heart_rate": 181,
                "lactate_4": 3.3,
                "lactate_8": 3.8,
                "cadence": 94,
                "rpe": 18,
                "feeling": "Zwaar",
            },
            {
                "step": 7,
                "length": 8,
                "intensity": 340,
                "rel_power": 4.3,
                "heart_rate": 190,
                "lactate_4": 6.4,
                "lactate_8": 7.5,
                "cadence": 92,
                "rpe": 19,
                "feeling": "Zeer zeer zwaar",
            },
        ],
    }



from typing import Optional

import pandas as pd
from pydantic import BaseModel, ConfigDict


class BaseMeasurement(BaseModel):
    lactate: float
    intensity: float
    heart_rate: float


class LactateTurningPoint(BaseMeasurement):
    pass


class ModDMax(BaseMeasurement):
    pass


class LogLog(BaseMeasurement):
    pass


class BaseLinePlus(BaseMeasurement):
    plus: float


class OBLA(BaseMeasurement):
    pass


class ThresholdEstimate(BaseMeasurement):
    pass


class LactateThresholdResults(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    clean_data: pd.DataFrame
    interpolated_data: pd.DataFrame
    ltp1: LactateTurningPoint | None = None
    ltp2: LactateTurningPoint | None = None
    mod_dmax: ModDMax | None = None
    loglog: LogLog | None = None
    baseline: BaseLinePlus | None = None
    obla_2: OBLA | None = None
    obla_4: OBLA | None = None
    lt1_estimate: ThresholdEstimate | None = None
    lt2_estimate: ThresholdEstimate | None = None

    def calc_lt1_lt2_estimates(self, lt1: Optional[float] = None, lt2: Optional[float] = None):
        """
        Calculate the LT1 and LT2 estimates. As a method so that interactively this can be updated.
        """
        from lactate_thresholds.methods import determine_threshold_estimate

        self.lt1_estimate = determine_threshold_estimate(self.interpolated_data, lt1, self.ltp1, self.loglog)
        self.lt2_estimate = determine_threshold_estimate(self.interpolated_data, lt2, self.ltp2, self.mod_dmax)



import logging

import pandas as pd

from lactate_thresholds import determine, methods, process


def test_interpolation(test_instances):
    df = pd.DataFrame.from_dict(test_instances["simple"])
    dfc = process.clean_data(df)
    dfi = methods.interpolate(dfc, include_baseline=False)
    logging.info(methods.determine_ltp(dfc, dfi))


def test_interpolation_watt_intensity(test_instances):
    df = pd.DataFrame.from_dict(test_instances["cycling1"])
    dfc = process.clean_data(df, lactate_col="lactate_8")
    dfi = methods.interpolate(dfc, include_baseline=False)
    logging.info(methods.determine_ltp(dfc, dfi))

    # do the same for cycling2
    df = pd.DataFrame.from_dict(test_instances["cycling2"])
    dfc = process.clean_data(df, lactate_col="lactate_8")
    dfi = methods.interpolate(dfc, include_baseline=False)
    logging.info(methods.determine_ltp(dfc, dfi))


def test_convenience_determine(test_instances):
    df = pd.DataFrame.from_dict(test_instances["cycling2"])
    determine(df, lactate_col="lactate_4")


def test_moddmax(test_instances):
    df = pd.DataFrame.from_dict(test_instances["cycling2"])
    dfc = process.clean_data(df, lactate_col="lactate_8")
    dfi = methods.interpolate(dfc, include_baseline=False)
    logging.info(methods.determine_mod_dmax(dfc, dfi))


def test_loglog(test_instances):
    df = pd.DataFrame.from_dict(test_instances["cycling2"])
    dfc = process.clean_data(df, lactate_col="lactate_8")
    di = methods.interpolate(dfc, include_baseline=False)
    logging.info(methods.determine_loglog(dfc, di))


def test_baseline_plus(test_instances):
    df = pd.DataFrame.from_dict(test_instances["cycling2"])
    dfc = process.clean_data(df, lactate_col="lactate_8")
    di = methods.interpolate(dfc, include_baseline=False)
    logging.info(methods.determine_baseline(dfc, di, 0.5))


def test_obla(test_instances):
    df = pd.DataFrame.from_dict(test_instances["cycling2"])
    dfc = process.clean_data(df, lactate_col="lactate_8")
    di = methods.interpolate(dfc, include_baseline=False)
    logging.info(methods.determine_obla(di, 4.0))
    logging.info(methods.determine_obla(di, 2.0))



import altair as alt
import pandas as pd

from lactate_thresholds.types import LactateThresholdResults


def lactate_intensity_plot(x: LactateThresholdResults, show_fit_line: bool = True):
    clean_data = x.clean_data[x.clean_data["intensity"] > 0]
    interpolated_data = x.interpolated_data

    base = (
        alt.Chart(clean_data)
        .encode(
            x=alt.X(
                "intensity:Q",
                title="Intensity",
                scale=alt.Scale(
                    domain=[
                        clean_data["intensity"].min() - 1,
                        clean_data["intensity"].max() + 1,
                    ]
                ),
            ),
            y=alt.Y("lactate:Q", title="Lactate"),
        )
        .properties(width=800, height=600)
    )

    points_orig = base.mark_point(color="grey", opacity=0.3).properties(title="Lactate Intensity Plot")
    line_orig = base.mark_line(color="grey", opacity=0.3)

    if show_fit_line:
        line_interpolated = alt.Chart(interpolated_data).mark_line().encode(x="intensity:Q", y="lactate:Q")

    # Add thresholds with shapes and colors
    threshold_data = []
    shapes = {
        "ltp1": ("circle", "#FF6347"),  # tomato
        "ltp2": ("square", "#4682B4"),  # steelblue
        "mod_dmax": ("diamond", "#32CD32"),  # limegreen
        "loglog": ("cross", "#FFA500"),  # orange
        "obla_2": ("triangle-up", "#8A2BE2"),  # blueviolet
        "obla_4": ("triangle-down", "#8A2BE2"),  # blueviolet
        "lt1_estimate": (
            "M0,.5L.6,.8L.5,.1L1,-.3L.3,-.4L0,-1L-.3,-.4L-1,-.3L-.5,.1L-.6,.8L0,.5Z",
            "#FFD700",
        ),  # gold
        "lt2_estimate": (
            "M0,.5L.6,.8L.5,.1L1,-.3L.3,-.4L0,-1L-.3,-.4L-1,-.3L-.5,.1L-.6,.8L0,.5Z",
            "#FFD700",
        ),  # gold
    }

    for key, (shape, color) in shapes.items():
        r = getattr(x, key)
        if r is not None:
            threshold_data.append(
                {
                    "intensity": r.intensity,
                    "lactate": r.lactate,
                    "threshold": key,
                    "shape": shape,
                    "color": color,
                }
            )

    threshold_df = pd.DataFrame(threshold_data)

    thresholds = (
        alt.Chart(threshold_df)
        .mark_point(size=150, filled=True, strokeOpacity=1, fillOpacity=1)
        .encode(
            x="intensity:Q",
            y="lactate:Q",
            shape=alt.Shape(
                "threshold:N",
                scale=alt.Scale(
                    domain=list(shapes.keys()),
                    range=[shapes[key][0] for key in shapes.keys()],
                ),
            ),
            color=alt.Color(
                "threshold:N",
                scale=alt.Scale(
                    domain=list(shapes.keys()),
                    range=[shapes[key][1] for key in shapes.keys()],
                ),
            ),
        )
    )

    # Add interactive selection tied to interpolated data
    nearest = alt.selection_point(nearest=True, on="mouseover", fields=["intensity"], empty=False)

    selectors = (
        alt.Chart(interpolated_data).mark_point().encode(x="intensity:Q", opacity=alt.value(0)).add_params(nearest)
    )

    points = (
        alt.Chart(interpolated_data)
        .mark_point(size=50, color="red")
        .encode(
            x="intensity:Q",
            y="lactate:Q",
            opacity=alt.condition(nearest, alt.value(1), alt.value(0)),
        )
    )

    rules = (
        alt.Chart(interpolated_data)
        .mark_rule(color="gray")
        .encode(x="intensity:Q")
        .transform_filter(nearest)
        .properties(
            height=600  # Ensure the rule spans the full height of the graph
        )
    )

    vertical_dotted_line = (
        alt.Chart(pd.DataFrame({"lactate": [x.baseline.lactate]}))
        .mark_rule(strokeDash=[5, 5], color="purple")
        .encode(y=alt.Y("lactate:Q"))
        .properties(width=800, height=600)
    )

    layers = [
        points_orig,
        line_orig,
        vertical_dotted_line,
        thresholds,
        selectors,
        points,
        rules,
    ]

    if show_fit_line:
        layers.append(line_interpolated)

    chart = (
        alt.layer(*layers)
        .interactive()
        .encode(
            tooltip=[
                alt.Tooltip("intensity:Q", format=".1f"),
                alt.Tooltip("lactate:Q", format=".1f"),
                alt.Tooltip("heart_rate:Q", format=".0f"),
            ]
        )
    )

    return chart


def heart_rate_intensity_plot(x: LactateThresholdResults, show_fit_line: bool = True):
    clean_data = x.clean_data[x.clean_data["intensity"] > 0]
    interpolated_data = x.interpolated_data

    base = (
        alt.Chart(clean_data)
        .encode(
            x=alt.X(
                "intensity:Q",
                title="Intensity",
                scale=alt.Scale(
                    domain=[
                        clean_data["intensity"].min() - 1,
                        clean_data["intensity"].max() + 1,
                    ]
                ),
            ),
            y=alt.Y(
                "heart_rate:Q",
                title="Heart Rate (bpm)",
                scale=alt.Scale(
                    domain=[
                        clean_data["heart_rate"].min() - 10,
                        clean_data["heart_rate"].max() + 10,
                    ]
                ),
            ),
        )
        .properties(width=800, height=600)
    )

    points_orig = base.mark_point(color="grey", opacity=0.3).properties(title="Heart Rate Intensity Plot")
    line_orig = base.mark_line(color="grey", opacity=0.3)

    if show_fit_line:
        line_interpolated = alt.Chart(interpolated_data).mark_line().encode(x="intensity:Q", y="heart_rate:Q")

    # Add thresholds with shapes and colors
    threshold_data = []
    shapes = {
        "ltp1": ("circle", "#FF6347"),  # tomato
        "ltp2": ("square", "#4682B4"),  # steelblue
        "mod_dmax": ("diamond", "#32CD32"),  # limegreen
        "loglog": ("cross", "#FFA500"),  # orange
        "obla_2": ("triangle-up", "#8A2BE2"),  # blueviolet
        "obla_4": ("triangle-down", "#8A2BE2"),  # blueviolet
        "lt1_estimate": (
            "M0,.5L.6,.8L.5,.1L1,-.3L.3,-.4L0,-1L-.3,-.4L-1,-.3L-.5,.1L-.6,.8L0,.5Z",
            "#FFD700",
        ),  # gold
        "lt2_estimate": (
            "M0,.5L.6,.8L.5,.1L1,-.3L.3,-.4L0,-1L-.3,-.4L-1,-.3L-.5,.1L-.6,.8L0,.5Z",
            "#FFD700",
        ),  # gold
    }

    for key, (shape, color) in shapes.items():
        r = getattr(x, key)
        if r is not None:
            threshold_data.append(
                {
                    "intensity": r.intensity,
                    "heart_rate": r.heart_rate,
                    "threshold": key,
                    "shape": shape,
                    "color": color,
                }
            )

    threshold_df = pd.DataFrame(threshold_data)

    thresholds = (
        alt.Chart(threshold_df)
        .mark_point(size=150, filled=True, strokeOpacity=1, fillOpacity=1)
        .encode(
            x="intensity:Q",
            y="heart_rate:Q",
            shape=alt.Shape(
                "threshold:N",
                scale=alt.Scale(
                    domain=list(shapes.keys()),
                    range=[shapes[key][0] for key in shapes.keys()],
                ),
            ),
            color=alt.Color(
                "threshold:N",
                scale=alt.Scale(
                    domain=list(shapes.keys()),
                    range=[shapes[key][1] for key in shapes.keys()],
                ),
            ),
        )
    )

    # Add interactive selection tied to interpolated data
    nearest = alt.selection_point(nearest=True, on="mouseover", fields=["intensity"], empty=False)

    selectors = (
        alt.Chart(interpolated_data).mark_point().encode(x="intensity:Q", opacity=alt.value(0)).add_params(nearest)
    )

    points = (
        alt.Chart(interpolated_data)
        .mark_point(size=50, color="red")
        .encode(
            x="intensity:Q",
            y="heart_rate:Q",
            opacity=alt.condition(nearest, alt.value(1), alt.value(0)),
        )
    )

    rules = (
        alt.Chart(interpolated_data)
        .mark_rule(color="gray")
        .encode(x="intensity:Q")
        .transform_filter(nearest)
        .properties(
            height=600  # Ensure the rule spans the full height of the graph
        )
    )

    layers = [
        points_orig,
        line_orig,
        thresholds,
        selectors,
        points,
        rules,
    ]

    if show_fit_line:
        layers.append(line_interpolated)

    chart = (
        alt.layer(*layers)
        .interactive()
        .encode(
            tooltip=[
                alt.Tooltip("intensity:Q", format=".1f"),
                alt.Tooltip("heart_rate:Q", format=".0f"),
            ]
        )
    )

    return chart



import pandas as pd


def example_data_cycling():
    return pd.DataFrame.from_dict(
        [
            {
                "step": 1,
                "length": 8,
                "intensity": 100,
                "rel_power": 1.3,
                "heart_rate": 113,
                "lactate_4": 1.0,
                "lactate_8": 1.0,
                "cadence": 102,
                "rpe": 6,
            },
            {
                "step": 2,
                "length": 8,
                "intensity": 140,
                "rel_power": 1.8,
                "heart_rate": 126,
                "lactate_4": 1.0,
                "lactate_8": 1.0,
                "cadence": 100,
                "rpe": 7,
            },
            {
                "step": 3,
                "length": 8,
                "intensity": 180,
                "rel_power": 2.3,
                "heart_rate": 137,
                "lactate_4": 0.9,
                "lactate_8": 0.9,
                "cadence": 100,
                "rpe": 10,
            },
            {
                "step": 4,
                "length": 8,
                "intensity": 220,
                "rel_power": 2.8,
                "heart_rate": 151,
                "lactate_4": 1.0,
                "lactate_8": 1.0,
                "cadence": 98,
                "rpe": 12,
            },
            {
                "step": 5,
                "length": 8,
                "intensity": 260,
                "rel_power": 3.3,
                "heart_rate": 168,
                "lactate_4": 1.9,
                "lactate_8": 1.9,
                "cadence": 98,
                "rpe": 16,
            },
            {
                "step": 6,
                "length": 8,
                "intensity": 300,
                "rel_power": 3.8,
                "heart_rate": 181,
                "lactate_4": 3.3,
                "lactate_8": 3.8,
                "cadence": 94,
                "rpe": 18,
            },
            {
                "step": 7,
                "length": 8,
                "intensity": 340,
                "rel_power": 4.3,
                "heart_rate": 190,
                "lactate_4": 6.4,
                "lactate_8": 7.5,
                "cadence": 92,
                "rpe": 19,
            },
        ]
    )



import pandas as pd

from lactate_thresholds.types import LactateThresholdResults


def seiler_3_zones(res: LactateThresholdResults) -> pd.DataFrame:
    """Determine Seiler 3-zone training zones based on LT1 and LT2.

    Args:
        lt1 (ThresholdEstimate): Lactate threshold 1 intensity.
        lt2 (ThresholdEstimate): Lactate threshold 2 intensity.

    Returns:
        pd.DataFrame: DataFrame with Seiler 3-zone training zones, including focus and usage.
    """

    from lactate_thresholds.utils import get_heart_rate_interpolated

    zone1_intensity = f"{0:.2f} - {res.lt1_estimate.intensity:.2f}"
    zone1_heartrate = f"up to {get_heart_rate_interpolated(res.interpolated_data, res.lt1_estimate.intensity):.0f}"

    zone2_intensity = f"{res.lt1_estimate.intensity:.2f} - {res.lt2_estimate.intensity:.2f}"
    zone2_heartrate = (
        f"{get_heart_rate_interpolated(res.interpolated_data, res.lt1_estimate.intensity):.0f} - "
        f"{get_heart_rate_interpolated(res.interpolated_data, res.lt2_estimate.intensity):.0f}"
    )

    zone3_intensity = f"{res.lt2_estimate.intensity:.2f} - max"
    zone3_heartrate = f"{get_heart_rate_interpolated(res.interpolated_data, res.lt2_estimate.intensity):.0f} - max"

    zones = pd.DataFrame(
        {
            "zone": ["Zone 1", "Zone 2", "Zone 3"],
            "intensity": [zone1_intensity, zone2_intensity, zone3_intensity],
            "heart_rate": [zone1_heartrate, zone2_heartrate, zone3_heartrate],
            "focus": [
                "Recovery, building an aerobic foundation.",
                "Moderate aerobic work; a gray zone with limited efficiency for endurance adaptations.",
                "Enhancing anaerobic threshold and lactate tolerance.",
            ],
        }
    )

    return zones


def seiler_5_zones(res: LactateThresholdResults) -> pd.DataFrame:
    """Determine 5-zone training zones based on LT1 and LT2.

    Args:
        res (LactateThresholdResults): Object containing LT1 and LT2 estimates and interpolated data.

    Returns:
        pd.DataFrame: DataFrame with 5-zone training zones, including focus and usage.
    """

    from lactate_thresholds.utils import get_heart_rate_interpolated

    lt1_intensity = res.lt1_estimate.intensity
    lt2_intensity = res.lt2_estimate.intensity

    zone1_intensity = f"0 - {0.98 * lt1_intensity:.2f}"
    zone2_intensity = f"{0.98 * lt1_intensity:.2f} - {1.02 * lt1_intensity:.2f}"
    zone3_intensity = f"{1.02 * lt1_intensity:.2f} - {0.98 * lt2_intensity:.2f}"
    zone4_intensity = f"{0.98 * lt2_intensity:.2f} - {1.02 * lt2_intensity:.2f}"
    zone5_intensity = f"{1.02 * lt2_intensity:.2f} - max"

    zone1_heartrate = f"up to {get_heart_rate_interpolated(res.interpolated_data, 0.98 * lt1_intensity):.0f}"
    zone2_heartrate = (
        f"{get_heart_rate_interpolated(res.interpolated_data, 0.98 * lt1_intensity):.0f} - "
        f"{get_heart_rate_interpolated(res.interpolated_data, 1.02 * lt1_intensity):.0f}"
    )
    zone3_heartrate = (
        f"{get_heart_rate_interpolated(res.interpolated_data, 1.02 * lt1_intensity):.0f} - "
        f"{get_heart_rate_interpolated(res.interpolated_data, 0.98 * lt2_intensity):.0f}"
    )
    zone4_heartrate = (
        f"{get_heart_rate_interpolated(res.interpolated_data, 0.98 * lt2_intensity):.0f} - "
        f"{get_heart_rate_interpolated(res.interpolated_data, 1.02 * lt2_intensity):.0f}"
    )
    zone5_heartrate = f"{get_heart_rate_interpolated(res.interpolated_data, 1.02 * lt2_intensity):.0f} - max"

    zones = pd.DataFrame(
        {
            "zone": ["Zone 1", "Zone 2", "Zone 3", "Zone 4", "Zone 5"],
            "intensity": [
                zone1_intensity,
                zone2_intensity,
                zone3_intensity,
                zone4_intensity,
                zone5_intensity,
            ],
            "heart_rate": [
                zone1_heartrate,
                zone2_heartrate,
                zone3_heartrate,
                zone4_heartrate,
                zone5_heartrate,
            ],
            "focus": [
                "Recovery, building an aerobic foundation.",
                "Aerobic base building and improving fat utilization.",
                "Aerobic endurance and muscular efficiency.",
                "Improving lactate tolerance and anaerobic threshold.",
                "Anaerobic capacity, speed, and power.",
            ],
        }
    )

    return zones


# friel references
# https://web.archive.org/web/20241212065559/https://www.trainingbible.com/joesblog/2009/11/quick-guide-to-setting-zones.html


def friel_7_zones_running(res: LactateThresholdResults) -> pd.DataFrame:
    """Determine Friel's 7-zone training zones based on LT (Lactate Threshold).

    Args:
        res (LactateThresholdResults): Object containing LT2 estimate and interpolated data.

    Returns:
        pd.DataFrame: DataFrame with Friel's 7-zone training zones.
    """
    from lactate_thresholds.utils import (
        get_intensity_based_on_heartrate_interpolated,
    )

    lt2_heart_rate = res.lt2_estimate.heart_rate

    # Define zone boundaries based on LT2
    hr_zone1 = [0, 0.85 * lt2_heart_rate]
    hr_zone2 = [0.85 * lt2_heart_rate, 0.89 * lt2_heart_rate]
    hr_zone3 = [0.90 * lt2_heart_rate, 0.94 * lt2_heart_rate]
    hr_zone4 = [0.95 * lt2_heart_rate, 0.99 * lt2_heart_rate]
    hr_zone5a = [1.00 * lt2_heart_rate, 1.02 * lt2_heart_rate]
    hr_zone5b = [1.03 * lt2_heart_rate, 1.06 * lt2_heart_rate]
    hr_zone5c = [1.06 * lt2_heart_rate, res.clean_data.heart_rate.max()]

    # Create the zones DataFrame
    zones = pd.DataFrame(
        {
            "zone": [
                "Zone 1. Recovery",
                "Zone 2. Aerobic",
                "Zone 3. Tempo",
                "Zone 4. SubThreshold",
                "Zone 5a. VO2 SuperThreshold",
                "Zone 5b. Aerobic Capacity",
                "Zone 5c. Anaerobic Capacity",
            ],
            "intensity": [
                f"up to {get_intensity_based_on_heartrate_interpolated(res.interpolated_data, hr_zone1[1]):.2f}",
                f"{get_intensity_based_on_heartrate_interpolated(res.interpolated_data, hr_zone2[0]):.2f} - "
                f"{get_intensity_based_on_heartrate_interpolated(res.interpolated_data, hr_zone2[1]):.2f}",
                f"{get_intensity_based_on_heartrate_interpolated(res.interpolated_data, hr_zone3[0]):.2f} - "
                f"{get_intensity_based_on_heartrate_interpolated(res.interpolated_data, hr_zone3[1]):.2f}",
                f"{get_intensity_based_on_heartrate_interpolated(res.interpolated_data, hr_zone4[0]):.2f} - "
                f"{get_intensity_based_on_heartrate_interpolated(res.interpolated_data, hr_zone4[1]):.2f}",
                f"{get_intensity_based_on_heartrate_interpolated(res.interpolated_data, hr_zone5a[0]):.2f} - "
                f"{get_intensity_based_on_heartrate_interpolated(res.interpolated_data, hr_zone5a[1]):.2f}",
                f"{get_intensity_based_on_heartrate_interpolated(res.interpolated_data, hr_zone5b[0]):.2f} - "
                f"{get_intensity_based_on_heartrate_interpolated(res.interpolated_data, hr_zone5b[1]):.2f}",
                f"more than {get_intensity_based_on_heartrate_interpolated(res.interpolated_data, hr_zone5c[0]):.2f}",
            ],
            "heart_rate": [
                f"{hr_zone1[0]:.0f} - {hr_zone1[1]:.0f}",
                f"{hr_zone2[0]:.0f} - {hr_zone2[1]:.0f}",
                f"{hr_zone3[0]:.0f} - {hr_zone3[1]:.0f}",
                f"{hr_zone4[0]:.0f} - {hr_zone4[1]:.0f}",
                f"{hr_zone5a[0]:.0f} - {hr_zone5a[1]:.0f}",
                f"{hr_zone5b[0]:.0f} - {hr_zone5b[1]:.0f}",
                f"{hr_zone5c[0]:.0f} - max",
            ],
            "focus": [
                "Active recovery.",
                "Aerobic endurance.",
                "Building aerobic capacity and stamina.",
                "Threshold effort.",
                "Improving VO2 max.",
                "Anaerobic capacity.",
                "Peak power output.",
            ],
        }
    )

    return zones



import pandas as pd

from lactate_thresholds.methods import (
    determine_baseline,
    determine_loglog,
    determine_ltp,
    determine_mod_dmax,
    determine_obla,
    interpolate,
)
from lactate_thresholds.types import LactateThresholdResults


def clean_data(
    df: pd.DataFrame,
    step_col: str = "step",
    length_col: str = "length",
    intensity_col: str = "intensity",
    lactate_col: str = "lactate",
    heart_rate_col: str = "heart_rate",
) -> pd.DataFrame:
    ## if df not a dataframe raise valueerror
    if not isinstance(df, pd.DataFrame):
        raise ValueError("Input is not a DataFrame")

    df_clean = df.copy()
    df_clean = df_clean.rename(
        columns={
            step_col: "step",
            length_col: "length",
            intensity_col: "intensity",
            lactate_col: "lactate",
            heart_rate_col: "heart_rate",
        }
    )
    df_clean = df_clean[["step", "length", "intensity", "lactate", "heart_rate"]]

    ## iterate over columns and assert that all entries in the column are numeric
    for col in df_clean.columns:
        if not pd.api.types.is_numeric_dtype(df_clean[col]):
            raise ValueError(f"Column '{col}' is not numeric / contains nonnumeric values")

    return df_clean


def determine(
    df: pd.DataFrame,
    step_col: str = "step",
    length_col: str = "length",
    intensity_col: str = "intensity",
    lactate_col: str = "lactate",
    heart_rate_col: str = "heart_rate",
    include_baseline=False,
) -> LactateThresholdResults:
    dfc = clean_data(df, step_col, length_col, intensity_col, lactate_col, heart_rate_col)
    dfi = interpolate(dfc, include_baseline=include_baseline)

    res = LactateThresholdResults(clean_data=dfc, interpolated_data=dfi)
    res.ltp1, res.ltp2 = determine_ltp(dfc, dfi)
    res.mod_dmax = determine_mod_dmax(dfc, dfi)
    res.loglog = determine_loglog(dfc, dfi)
    res.obla_2 = determine_obla(dfi, 2)
    res.obla_4 = determine_obla(dfi, 4)
    res.baseline = determine_baseline(dfc, dfi, 0)
    res.calc_lt1_lt2_estimates()

    return res



from lactate_thresholds.plot import lactate_intensity_plot
from lactate_thresholds.process import clean_data, determine, interpolate

__all__ = ["lactate_intensity_plot", "clean_data", "determine", "interpolate"]



import pandas as pd
import pytest

from lactate_thresholds import process


def test_lactate_data_ingestion_simple(test_instances):
    df = pd.DataFrame.from_dict(test_instances["simple"])
    df2 = process.clean_data(df)
    assert set(df2.columns) == set(["step", "length", "intensity", "lactate", "heart_rate"])


def test_lactate_data_ingestion_rename(test_instances):
    df = pd.DataFrame.from_dict(test_instances["differently_named_cols"])
    df2 = process.clean_data(df, heart_rate_col="heartrate", length_col="step_length")
    assert set(df2.columns) == set(["step", "length", "intensity", "lactate", "heart_rate"])

    df_cycling = pd.DataFrame.from_dict(test_instances["cycling1"])
    df2_cycling = process.clean_data(df_cycling, lactate_col="lactate_8")
    assert set(df2_cycling.columns) == set(["step", "length", "intensity", "lactate", "heart_rate"])


def test_lactate_data_ingestion_non_numeric(test_instances):
    df = pd.DataFrame.from_dict(test_instances["non_numeric_vals"])
    with pytest.raises(ValueError):
        process.clean_data(df, heart_rate_col="heartrate")


def test_should_raise_err_if_not_dataframe(test_instances):
    with pytest.raises(ValueError):
        process.clean_data(test_instances["non_numeric_vals"])



import pandas as pd

from lactate_thresholds import methods, process


def test_interpolation(test_instances):
    ld = pd.DataFrame.from_dict(test_instances["simple"])
    df = process.clean_data(ld)
    idf = methods.interpolate(df)
    # assert idf has cols 'heart_rate', 'lactate' and 'intensity'
    assert "heart_rate" in idf.columns
    assert "lactate" in idf.columns
    assert "intensity" in idf.columns



import numpy as np
import pandas as pd
import statsmodels.api as sm


def get_heart_rate(df_clean: pd.DataFrame, intensity_values: np.array) -> np.array:
    df_clean = df_clean.iloc[1:]

    X = sm.add_constant(df_clean["intensity"])
    y = df_clean["heart_rate"]
    model_heart_rate = sm.OLS(y, X).fit()

    new_data = pd.DataFrame({"const": 1, "intensity": intensity_values})
    return model_heart_rate.predict(new_data).round(0)


def get_heart_rate_interpolated(df_interpolated: pd.DataFrame, intensity: float) -> float:
    closest_intensity = df_interpolated.iloc[(df_interpolated["intensity"] - intensity).abs().argsort()[:1]]
    return closest_intensity["heart_rate"].values[0]


def get_lactate_interpolated(df_interpolated: pd.DataFrame, intensity: float) -> float:
    closest_intensity = df_interpolated.iloc[(df_interpolated["intensity"] - intensity).abs().argsort()[:1]]
    return closest_intensity["lactate"].values[0]


def get_intensity_interpolated(df_interpolated: pd.DataFrame, lactate: float) -> float:
    closest_lactate = df_interpolated.iloc[(df_interpolated["lactate"] - lactate).abs().argsort()[:1]]
    return closest_lactate["intensity"].values[0]


def get_intensity_based_on_heartrate_interpolated(df_interpolated: pd.DataFrame, heart_rate: float) -> float:
    closest_hr = df_interpolated.iloc[(df_interpolated["heart_rate"] - heart_rate).abs().argsort()[:1]]
    return closest_hr["intensity"].values[0]



import logging

import pandas as pd

from lactate_thresholds import determine
from lactate_thresholds.plot import heart_rate_intensity_plot, lactate_intensity_plot


def test_lactate_intensity_plot(test_instances, test_output_dir):
    df = pd.DataFrame.from_dict(test_instances["simple"])
    df2 = determine(df)
    logging.info(df2)
    chart = lactate_intensity_plot(df2)
    chart.save(f"{test_output_dir}/lactate_intensity_plot.html")


def test_lactate_intensity_plot2(test_instances, test_output_dir):
    df = pd.DataFrame.from_dict(test_instances["cycling2"])
    df2 = determine(df, lactate_col="lactate_8")
    chart = lactate_intensity_plot(df2)
    chart.save(f"{test_output_dir}/lactate_intensity_plot2.html")


def test_heartrate_intensity_plot(test_instances, test_output_dir):
    df = pd.DataFrame.from_dict(test_instances["simple"])
    df2 = determine(df)
    logging.info(df2)
    chart = heart_rate_intensity_plot(df2)
    chart.save(f"{test_output_dir}/heartrate_intensity_plot.html")



import logging
from typing import List, Optional

import numpy as np
import pandas as pd
import pwlf
import statsmodels.api as sm
from numpy.polynomial.polynomial import Polynomial
from scipy.optimize import curve_fit

from lactate_thresholds.types import (
    OBLA,
    BaseLinePlus,
    LactateTurningPoint,
    LogLog,
    ModDMax,
    ThresholdEstimate,
)
from lactate_thresholds.utils import (
    get_heart_rate,
    get_heart_rate_interpolated,
    get_intensity_interpolated,
    get_lactate_interpolated,
)


def interpolate(df: pd.DataFrame, interpolation_factor: float = 0.1, include_baseline: bool = True) -> pd.DataFrame:
    # Adjust baseline intensity
    if include_baseline and (df["intensity"] == 0).any():
        to_subtract = df.iloc[2]["intensity"] - df.iloc[1]["intensity"]
        df.loc[df["intensity"] == 0, "intensity"] = df.iloc[1]["intensity"] - to_subtract

    # Remove baseline if not included
    if not include_baseline and (df["intensity"] == 0).any():
        df = df[df["intensity"] != 0]

    # Sort the dataframe by intensity
    df = df.sort_values(by="intensity").reset_index(drop=True)

    def fit_and_predict(column):
        # Create polynomial features manually
        X = np.column_stack([df["intensity"] ** i for i in range(1, 4)])  # 3rd degree
        y = df[column]

        # Fit the statsmodels OLS model
        X = sm.add_constant(X)  # Add intercept term
        model = sm.OLS(y, X).fit()

        # Generate new intensity values for interpolation
        new_intensity = np.arange(df["intensity"].min(), df["intensity"].max(), interpolation_factor)
        new_X = np.column_stack([new_intensity**i for i in range(1, 4)])
        new_X = sm.add_constant(new_X)  # Add intercept term for prediction

        # Predict values for the new intensity data
        new_values = model.predict(new_X)
        return new_intensity, new_values

    # Interpolate lactate
    new_intensity, new_lactate = fit_and_predict("lactate")

    # Interpolate heartrate
    _, new_heartrate = fit_and_predict("heart_rate")

    # Combine interpolated values into a new DataFrame
    interpolated_df = pd.DataFrame(
        {
            "intensity": new_intensity,
            "lactate": new_lactate,
            "heart_rate": new_heartrate,
        }
    )

    return interpolated_df


def determine_ltp(
    data_clean: pd.DataFrame, data_interpolated: pd.DataFrame, n_breakpoints: int = 2
) -> List[LactateTurningPoint]:
    # Extract intensity and lactate data
    X = data_interpolated["intensity"].values
    y = data_interpolated["lactate"].values

    # Fit piecewise linear model
    pwlf_model = pwlf.PiecewiseLinFit(X, y)
    breakpoints = pwlf_model.fit(n_breakpoints + 1)  # n_segments = n_breakpoints + 1

    # Get breakpoint intensities and corresponding lactate values
    breakpoint_intensities = breakpoints[1:-1]  # Ignore first and last points

    lt1 = LactateTurningPoint(
        lactate=get_lactate_interpolated(data_interpolated, breakpoint_intensities[0]),
        intensity=breakpoint_intensities[0],
        heart_rate=get_heart_rate_interpolated(data_interpolated, breakpoint_intensities[0]),
    )
    lt2 = LactateTurningPoint(
        lactate=get_lactate_interpolated(data_interpolated, breakpoint_intensities[1]),
        intensity=breakpoint_intensities[1],
        heart_rate=get_heart_rate_interpolated(data_interpolated, breakpoint_intensities[1]),
    )

    return [lt1, lt2]


def determine_mod_dmax(data_clean: pd.DataFrame, data_interpolated: pd.DataFrame) -> ModDMax:
    if data_clean.empty or data_clean.iloc[0]["intensity"] == 0:
        data_dmax = data_clean.iloc[1:].copy()
    else:
        data_dmax = data_clean.copy()

    # Find the first rise in blood lactate greater than 0.4 mmol/L
    data_dmax["diffs"] = data_dmax["lactate"].diff().shift(-1)
    data_first_rise = data_dmax[data_dmax["diffs"] >= 0.4].head(1)

    if data_first_rise.empty:
        logging.warning("No first rise in blood lactate greater than 0.4 mmol/L found.")
        return None

    # Fit a 3rd degree polynomial
    def poly3(x, a, b, c, d):
        return a * x**3 + b * x**2 + c * x + d

    popt, _ = curve_fit(poly3, data_clean["intensity"], data_clean["lactate"])

    # Calculate the differences
    diff_lactate = data_dmax["lactate"].max() - data_first_rise["lactate"].values[0]
    diff_intensity = data_dmax["intensity"].max() - data_first_rise["intensity"].values[0]

    lin_beta = diff_lactate / diff_intensity

    # Find where the first derivative of the polynomial fit equals the slope of the line
    p = Polynomial([popt[3], popt[2], popt[1], popt[0]])
    dp = p.deriv()
    roots = (dp - lin_beta).roots()
    roots = roots[np.isreal(roots)].real
    roots = roots[roots > 0]

    max_intensity = data_dmax["intensity"].max()
    model_intensity = roots[roots <= max_intensity].max()
    model_lactate = poly3(model_intensity, *popt)

    # Workaround for unplausible estimations
    if model_lactate > 8:
        logging.warning("Estimated lactate value via ModDMax is higher than 8 mmol/L. Returning None.")
        return None

    return ModDMax(
        lactate=get_lactate_interpolated(data_interpolated, model_intensity),
        intensity=model_intensity,
        heart_rate=get_heart_rate(data_clean, [model_intensity])[0],
    )


def determine_loglog(data_clean: pd.DataFrame, data_interpolated: pd.DataFrame, loglog_restrainer=1):
    data_filtered = data_interpolated[data_interpolated["intensity"] > 0].copy()
    data_filtered["intensity"] = np.log(data_filtered["intensity"])
    data_filtered["lactate"] = np.log(data_filtered["lactate"])

    n = len(data_filtered)
    data_filtered = data_filtered.iloc[: int(loglog_restrainer * n)]

    piecewise_fit = pwlf.PiecewiseLinFit(data_filtered["intensity"].values, data_filtered["lactate"].values)
    breakpoints = piecewise_fit.fit(2)  # 1 breakpoint means 2 segments

    loglog_intensity = np.exp(breakpoints[1])
    lactate_interpolated = get_lactate_interpolated(data_interpolated, loglog_intensity)
    loglog_heart_rate = get_heart_rate_interpolated(data_interpolated, loglog_intensity)

    return LogLog(
        lactate=lactate_interpolated,
        intensity=loglog_intensity,
        heart_rate=loglog_heart_rate,
    )


def determine_baseline(
    data_clean: pd.DataFrame,
    data_interpolated: pd.DataFrame,
    plus: float = 0.5,
    perc_initial_values: float = 0.2,
) -> pd.DataFrame:
    baseline_window = int(len(data_interpolated) * perc_initial_values)
    bsln = data_interpolated["lactate"].iloc[:baseline_window].mean()
    bsln_plus = bsln + plus

    if bsln_plus > data_interpolated["lactate"].max() or bsln_plus < data_interpolated["lactate"].min():
        logging.warning(f"Baseline + {plus} is out of range.")
        return None

    bsln_plus_intensity = get_intensity_interpolated(data_interpolated, bsln_plus)
    return BaseLinePlus(
        lactate=bsln_plus,
        intensity=bsln_plus_intensity,
        heart_rate=get_heart_rate_interpolated(data_interpolated, bsln_plus_intensity),
        plus=plus,
    )


def determine_obla(data_interpolated: pd.DataFrame, obla_lactate: float) -> OBLA:
    obla_intensity = get_intensity_interpolated(data_interpolated, obla_lactate)
    return OBLA(
        lactate=obla_lactate,
        intensity=obla_intensity,
        heart_rate=get_heart_rate_interpolated(data_interpolated, obla_intensity),
    )


def determine_threshold_estimate(
    data_interpolated: pd.DataFrame, intensity: Optional[float] = None, *args
) -> ThresholdEstimate:
    if intensity is None:
        intensity = np.mean([arg.intensity for arg in args])

    return ThresholdEstimate(
        intensity=np.round(intensity, 1),
        lactate=np.round(get_lactate_interpolated(data_interpolated, intensity), 1),
        heart_rate=np.round(get_heart_rate_interpolated(data_interpolated, intensity), 0),
    )
